"""Scheduler API offers the possibility to schedule running of tasks within current application. 
It may serve as main entry point for application, as well as side support for application run.

It offers similar functionality to `crontab` utility in Linux, however implementation of this 
functionality is completely custom.


### Example

Write a decorator for your custom function and run.

```python
from tma_common import ScheduledRun

# custom function with decorator
@ScheduledRun(minute='13',hour='*/2',weekday='1-5')
def func1(scheduler,current_time):
    scheduler.logStatus()
```

Custom function `func1` takes two parameters. First `scheduler` is scheduler object. Usually we
do not need this object, but for debug purposes it may be useful. Second parameter `currentTime`
is `datetime` object native to Python.

Parameters for function decorator are described in section **Scheduler Parameters**.

**Very important**: custom function is executed asynchronously in separate thread, 
that means if you schedule events for every minute of day, and your task takes 10 minutes for example,
then execution time does not interfere with triggering of events. As a result, your application may at the 
times be executing simultaneously around 10 tasks, each one started a minute after another.


### Scheduler Parameters

Parameters for function decorator are completely equivalent to parameters of SchedulerConfig class constructor.
They are based on format of `crontab` utility values.

Scheduler offers scheduling events with various granularity, where all granularities has to be satisfied in order to fire event.

Following table shows what parameters are available for Scheduler.

Parameter | Description
--------- | ------------------------------------------------
minute    | minute of hour (0, 1, ...59). For example value '5' matches times 08:05, 09:05, 10:05, ....
hour      | hour of day (0, 1, ... 23)
day       | day of month (1, 2, ...., 31)
month     | month of year (1-January, 2-February, ....12-December)
year      | year
weekday   | day of week, values: 0-Sunday, 1-Monday, ..., 6-Saturday
yearday   | day of year, values: 0, 1, ...365. value 0 is Jan 1, value 50 is February 20, value 120 is May 1 in non-leap year but is also Apr 30 in leap year

These parameters, if defined for scheduler, has to be satisfied simultaneously in order to fire event. Default value is, than all values of given parameter are valid for firing the event.

Values for each parameter can be defined in multiple ways. Scheduler support not only scalars, but also list of scalars, intervals and periods.

Format | Meaning
------- | -------------------
'*'     |  all values of given parameter are acceptable, that means every minute, every hour, every day ... depending on parameter itself
'*/5'   | all fifths, every 5 minute, every 5 hours, etc
'2-7'   | values in range of 2 to 7. From 2nd to 7th minute, from 2nd to 7th day, ...
'10-50/10' | every 10th in the range from 10 to 50, namely it means 10, 20, 30, 40, 50. Every x-th is taken from start of interval, not from absolute value. That means '13-55/7' is expanded to values 13, 20, 27, 34, 41, 48, 55
'1,2,5,6' | list of explicit values. In context of weekdays these values mean Monday, Tuesday, Friday, Saturday.

**@ScheduledRun(minute='13',hour='*/2',weekday='1-5')**

This makes an event every working day (Monday to Friday) every two hours exactly on 13th minute. So for monday it will be scheduled for these times:

```
00:13
02:13
04:13
06:13
...
22:13
```

**@ScheduledRun(day='*/5')**

This makes and event every minute during whole day on every fifth day in the month. That means 1440 events during day (1 day = 1440 minutes). This is because minute and hour parameters are not set here, and default value is '*' what means 'each and every'.

Please not, that 'fifth' is derived from the beginning of month, that means the series of days may look like: Jan 5, Jan 10, Jan 15 ... Jan 30, Feb 5, Feb 10, Feb 15, Feb 20, Feb 25, Mar 5, ....

You may notice that between Jan 30 and Feb 5 are 6 days, and between Feb 25 and Mar 5 are 8 or 9 days depending on leap/no-leap year. If you need to run exactly every 5th day, you may use parameter `yearday` exemplified in next example.

**@ScheduledRun(yearday='*/13', minute='0',hour='12')**

This creates event every 13th day in the year at noon (once a day). That means series of dates may look like: Jan 12, Jan 25, Feb 7, Feb 20, Mar 5, Mar 18, Mar 31, Apr 13, ...

## Advanced

If you need even more precise control over firing events, just remember that you may place any logic between scheduler-event and execution of your task. Your logic may filter-out any unwanted times. In the most extreme case, you may just scheduler for triggering event every minute, and do you own filtering of times completely.

```python

@ScheduledRun()
def my_own_func(scheduler,currentDateTime):
    if currentDateTime.day==3 and currentDateTime.month==4:
        # do my task in the background
        pass
```

"""

__all__ = [
    "set_test_mode",
    "is_test_mode",
    "scheduled",
    "request_shutdown",
    "is_shutdown_requested",
    "ScheduledPlan",
    "SchedulerConfig"
]
from .scheduler import ScheduledRun, request_shutdown, SchedulerConfig, is_shutdown_requested
from .scheduled_plan import ScheduledPlan
from .run_mode import set_test_mode, is_test_mode

scheduled = ScheduledRun  # alias for convenience



from datetime import datetime, timedelta
from typing import List
import time


class ScheduledPlanError(RuntimeError):
    pass


class EventPatterns:
    APPLY_INCLUDE = 1
    APPLY_EXCLUDE = 2

    def __init__(self):
        self.apply_type = EventPatterns.APPLY_INCLUDE
        self.hours = []
        self.days = []

    def contains_data(self):
        if len(self.hours)==0 and len(self.days)==0:
            return False
        if len(self.hours)==0:
            self.hours = [a for a in range(24)]
        if len(self.days)==0:
            self.days = [a for a in range(7)]
        return True


class ScheduledPlan():
    WEEKDAYS = ['mo', 'tu', 'we', 'th', 'fr', 'sa', 'su']
    g_last_runs = {}
    g_period = 60  # seconds

    @staticmethod
    def to_int(p, default):
        try:
            hr = int(p)
            return hr
        except Exception:
            return default
        
    @staticmethod
    def set_period(seconds:int):
        periond = ScheduledPlan.to_int(seconds, 60)
        if periond < 10:
            periond = 10
        if period > 3600:
            period = 3600
        ScheduledPlan.g_period = periond
            
    def __init__(self, name):
        self.name = name
        self.clear_map()
        # minimum number of seconds from last run
        self.maxtime_from_last = 3600
        self.no_hours = True
        self.last_run = 0

    def clear_map(self):
        self._map = [[False]*24 for _ in range(7)]
        self.no_hours = True

    def produce_array_range(self, start, end, modulo):
        counter = 0
        arr = []
        start = start % modulo
        end = end % modulo
        while counter < modulo:
            arr.append(start)
            if start == end:
                break
            start = (start + 1) % modulo
            counter += 1
        return arr

    def find_weekday(self, p):
        def find_by_name(name):
            for idx,wd in enumerate(ScheduledPlan.WEEKDAYS):
                if name.startswith(wd):
                    return idx
            return None
        if p in ['*', 'everyday']:
            return [a for a in range(7)]
        if '-' in p:
            range_days = p.split('-')
            if len(range_days) != 2:
                raise ScheduledPlanError('Range in weekdays has to contain only two values joined by dash char.')
            idx1 = find_by_name(range_days[0])
            idx2 = find_by_name(range_days[1])
            if idx1 is not None and idx2 is not None:
                return self.produce_array_range(idx1, idx2, 7)
            else:
                raise ScheduledPlanError(f'Cannot be interpreted as weeday range: {p}')

        idx = find_by_name(p)
        if idx is not None:
            return [idx]
        return []

    def find_hour(self, p):
        if p in ['*', 'every']:
            return [a for a in range(24)]
        if '-' in p:
            range_hours = p.split('-')
            if len(range_hours) != 2:
                raise ScheduledPlanError('Range in hours has to contain only two values joined by dash char.')
            idx1 = ScheduledPlan.to_int(range_hours[0], 0)
            idx2 = ScheduledPlan.to_int(range_hours[1], 23)
            return self.produce_array_range(idx1, idx2, 24)
        hr = ScheduledPlan.to_int(p, None)
        if hr is not None:
            return [hr]
        return []

    def _copy_to_map(self, inc:EventPatterns):
        value = (inc.apply_type == EventPatterns.APPLY_INCLUDE)
        for d in inc.days:
            for h in inc.hours:
                self.no_hours = False
                self._map[d][h] = value

    def _get_scripts(self, text) -> List[EventPatterns]:
        data = []
        mode = 0
        paths = text.lower().split(',')
        for path in paths:
            print(f'Parsing path: "{path}"')
            curr = EventPatterns()
            data.append(curr)
            for p in [a for a in path.split(' ') if len(a)>0]:
                if p.startswith('excl'):
                    curr.apply_type = EventPatterns.APPLY_EXCLUDE
                elif p.startswith('incl'):
                    curr.apply_type = EventPatterns.APPLY_INCLUDE
                elif p == 'on':
                    mode = 1
                elif p == 'at':
                    mode = 2
                elif p == 'freq':
                    mode = 3
                elif mode == 3:
                    mode = 0
                    self.set_frequency(p)
                elif mode == 1:
                    wd = self.find_weekday(p)
                    curr.days.extend(wd)
                elif mode == 2:
                    hr = self.find_hour(p)
                    curr.hours.extend(hr)
        return data

    def set_frequency(self, text):
        lengths = {
            'hour': 60*60,
            'day': 24*60*60,
            'week': 7*24*60*60
        }
        if text is None or not isinstance(text,str):
            raise ScheduledPlanError(f'Input for set_frequency has to be string')
        format_error = 'Argument for frequency has to be <number>/<interval>, where "number" is integer value and "interval" is from "hour", "day", "week".'
        p = text.split('/')
        if len(p)!=2:
            raise ScheduledPlanError(format_error)
        number = ScheduledPlan.to_int(p[0],-1)
        interval = p[1].lower()
        if interval not in ['day', 'hour', 'week']:
            raise ScheduledPlanError(format_error)
        if number < 1 or number > lengths[interval]:
            raise ScheduledPlanError(f'Number has to be from range <1,{lengths[interval]}) for interval "{interval}".')
        self.maxtime_from_last = int(lengths[interval]/number)
        


    def analyze(self, text):
        self.clear_map()
        data = self._get_scripts(text)
        for inc in data:
            if not inc.contains_data():
                continue
            self._copy_to_map(inc)

    def can_run(self, current:datetime):
        weekday = current.weekday()
        hour = current.hour

        if self.no_hours:
            hour_is_suitable = (self.maxtime_from_last is not None)
        else:
            hour_is_suitable = self._map[weekday][hour]

        if not hour_is_suitable:
            return False

        # if we are interested in frequency and have all info available
        if self.maxtime_from_last is not None and self.last_run is not None:
            # if last run was less than allowed time frame, do not allow
            from_last = current.timestamp() - self.last_run
            if from_last < self.maxtime_from_last:
                return False

        self.last_run = current.timestamp()
        return True
        

    def format_map(self):
        s = ''
        for day in self._map:
            for h in day:
                if h:
                    s += 'x'
                else:
                    s += '-'
            s += '\n'
        return s

if __name__=='__main__':
    tp = ScheduledPlan('plan1')
    tp.analyze('at 3 4 5 on mon tue fri sun, at 6 7 on tue, exclude at 4 on tue')
    print(tp.format_map())

    tp.analyze('on mon tu we fr th')
    print(tp.format_map())

    tp.analyze('at 9 10 on mon-fr, at 11 on sa su')
    print(tp.format_map())


# import json
import datetime as dt
import time
import threading
import logging
import signal
import sys

from .run_mode import is_test_mode
from .scheduled_plan import ScheduledPlan


logger = logging.getLogger(__name__)

# Global shutdown flag for graceful termination
_shutdown_requested = False
_shutdown_lock = threading.Lock()

def _signal_handler(signum, frame):
    """Handle termination signals by setting shutdown flag."""
    global _shutdown_requested
    with _shutdown_lock:
        _shutdown_requested = True
    logger.info(f"Received signal {signum}, initiating graceful shutdown...")

def _setup_signal_handlers():
    """Setup signal handlers for graceful shutdown."""
    # Handle common termination signals
    if hasattr(signal, 'SIGTERM'):
        signal.signal(signal.SIGTERM, _signal_handler)
    if hasattr(signal, 'SIGINT'):  # Ctrl+C
        signal.signal(signal.SIGINT, _signal_handler)
    # On Windows, handle SIGBREAK (Ctrl+Break)
    if sys.platform == 'win32' and hasattr(signal, 'SIGBREAK'):
        signal.signal(signal.SIGBREAK, _signal_handler)

def _is_shutdown_requested():
    """Check if shutdown has been requested."""
    with _shutdown_lock:
        return _shutdown_requested

def request_shutdown():
    """Manually request shutdown of all scheduler loops."""
    global _shutdown_requested
    with _shutdown_lock:
        _shutdown_requested = True
    logger.info("Manual shutdown requested for scheduler...")

# Initialize signal handlers
_setup_signal_handlers()

#
# safe conversion to integer from string
#
def to_int(text):
    try:
        val = int(text)
        return val, True
    except ValueError:
        return None, False


#
# count of days in year before given month
# 0 - January, 11 - December
#
yeardaycount = [0, 31, 59, 90, 120, 151, 181, 212, 243, 273, 304, 334, 365]


#
# range of values for event
# used only internally
#
class MomentRange:
    def __init__(self, text=None):
        self.set_range_text(text)
        self.range_from = 0
        self.range_to = 3000

    def dictionary(self):
        return {'from': self.range_from, 'to': self.range_to}

    def is_active(self, a):
        return a >= self.range_from and a <= self.range_to

    def set_range_text(self, text):
        va = vb = False
        if '-' in text:
            parts = text.split('-')
            if len(parts) != 2:
                logger.info(f'Too many parts of range in string {text}')
            self.range_from, va = to_int(parts[0])
            if not va:
                logger.info(f'Value "{parts[0]}" is not number')
                self.range_from = 0
            self.range_to, vb = to_int(parts[1])
            if not vb:
                logger.info(f'Value "{parts[1]}" is not number')
                self.range_to = 3000
        elif text == '*':
            self.range_from = 0
            self.range_to = 3000
        else:
            self.range_from, va = to_int(text)
            if not va:
                logger.info(f'Invalid number "{text}"')
                self.range_from = 0
                self.range_from = 3000
            else:
                self.range_to = self.range_from


#
# Period for scheduling event
# Used only internally
#
class MomentPeriod:
    def __init__(self, text):
        self.base = None
        self.period = 1
        if '/' in text:
            a = text.split('/')
            self.base = MomentRange(a[0])
            self.set_period_text(a[1])
        else:
            self.base = MomentRange(text)

    def dictionary(self):
        return {
            'range': self.base.dictionary(),
            'period': self.period
        }

    def is_active(self, num):
        return self.base.is_active(num) and \
               (num - self.base.range_from) % self.period == 0

    def set_period_text(self, text):
        self.period, va = to_int(text)
        if not va:
            self.period = 1
            logger.info(f'Invalid period {text}')


#
# List of values for scheduling event
# used only internally
#
class MomentList:
    def __init__(self, text):
        self.moments = []
        if ',' in text:
            parts = text.split(',')
            for p in parts:
                self.append_text(p)
        else:
            self.append_text(text)

    def dictionary(self):
        return [a.dictionary() for a in self.moments]

    def append_text(self, text):
        if '/' in text:
            self.moments.append(MomentPeriod(text))
        else:
            self.moments.append(MomentRange(text))

    def is_active(self, text):
        n, na = to_int(text)
        if not na:
            return False
        for m in self.moments:
            if m.is_active(n):
                return True
        return False


#
# Main class for scheduler
#
class SchedulerConfig:
    def __init__(self, minute=None, hour=None, weekday=None, day=None,
                 month=None, year=None, yearday=None):
        self.minute = self.text_to_numbers(minute or '*')
        self.hour = self.text_to_numbers(hour or '*')
        self.weekday = self.text_to_numbers(weekday or '*')
        self.day = self.text_to_numbers(day or '*')
        self.month = self.text_to_numbers(month or '*')
        self.year = self.text_to_numbers(year or '*')
        self.yearday = self.text_to_numbers(yearday or '*')

    def dictionary(self):
        value = {}
        if self.minute is not None:
            value['minute'] = self.minute.dictionary()
        if self.hour is not None:
            value['hour'] = self.hour.dictionary()
        if self.weekday is not None:
            value['weekday'] = self.weekday.dictionary()
        if self.day is not None:
            value['day'] = self.day.dictionary()
        if self.month is not None:
            value['month'] = self.month.dictionary()
        if self.year is not None:
            value['year'] = self.year.dictionary()
        if self.yearday is not None:
            value['yearday'] = self.yearday.dictionary()
        return value

    def text_to_numbers(self, text):
        if text is None or text == '*':
            return None
        if ',' in text:
            return MomentList(text)
        elif '/' in text:
            return MomentPeriod(text)
        else:
            return MomentRange(text)

    def next_datetime(self):
        d = dt.datetime.now()
        for a in range(0, 2):
            at = self.active_on_day(d)
            if at:
                return at
            d = dt.datetime(d.year, d.month, d.day, 0, 0) \
                + dt.timedelta(days=1)
        return None

    def day_of_year(self, d):
        leap = 0
        if d.year % 4 == 0 and (d.year % 100 != 0
           or d.year % 400 == 0) and d.month > 2:
            leap = 1
        return yeardaycount[d.month - 1] + (d.day - 1) + leap

    #
    # Internal function - to check if there will be event triggered
    # in current day (parameter d) and if yes, then returns only future events
    def active_on_day(self, d, target=None, limit=None):
        if self.year is not None and not self.year.is_active(d.year):
            return None
        if self.month is not None and not self.month.is_active(d.month):
            return None
        if self.day is not None and not self.day.is_active(d.day):
            return None
        if self.weekday is not None \
           and not self.weekday.is_active(d.isoweekday() % 7):
            return None
        if self.yearday is not None \
           and not self.yearday.is_active(self.day_of_year(d)):
            return None
        for hc in range(d.hour, 23):
            if self.hour is None or self.hour.is_active(hc):
                sm = d.minute if hc == d.hour else 0
                for hm in range(sm, 59):
                    if self.minute is None or self.minute.is_active(hm):
                        if target is None:
                            return dt.datetime(d.year, d.month, d.day, hc, hm)
                        else:
                            if limit is None or len(target) + 1 <= limit:
                                target.append(dt.datetime(d.year, d.month,
                                                          d.day, hc, hm))
                            else:
                                return None
        return None

    # Retrieve next events from NOW till specified days in future
    # or till we reach maximum number of entries (whichever comes first)
    def next_events(self, daysInFuture=7, maxEntries=10):
        entries = []
        d = dt.datetime.now()
        for a in range(0, daysInFuture):
            self.active_on_day(d, target=entries, limit=maxEntries)
            if len(entries) >= maxEntries:
                return entries
            # we are interested only in events
            # in range of (current_hour_min, 23:59) for today
            # and in range (00:00, 23:59) for all other days
            d = dt.datetime(d.year, d.month, d.day, 0, 0) \
                + dt.timedelta(days=1)
        return entries

    # only for debug purposes
    def log_status(self, daysInFuture=7, maxEntries=10):
        print('==========next events=============')
        for e in self.next_events(daysInFuture=daysInFuture,
                                 maxEntries=maxEntries):
            print(f'          {e}')

    # run scheduler in non-blocking mode (in background thread)
    def run_async(self, func):
        if is_test_mode():
            print('Subprocess mode detected, scheduler will not be started.')
            return
        
        t = threading.Thread(target=self.run, args=(func,))
        t.daemon = True  # Make thread daemon so it exits when main exits
        t.start()

    # run scheduler in blocking mode
    def run(self, func):
        if is_test_mode():
            print('Subprocess mode detected, scheduler will not be started.')
            return
        
        lastfire = 0
        self.log_status()
        # schedule forever (until shutdown is requested)
        while not _is_shutdown_requested():
            n = self.next_datetime()
            if n is None:  # no future events in near future, wait 1 day
                logger.debug('Wait for 1 day')
                # Sleep in smaller chunks to check shutdown flag more frequently
                for _ in range(864):  # 864 * 100 = 86400 seconds (1 day)
                    if _is_shutdown_requested():
                        break
                    time.sleep(100)
            else:  # some event in near future, wait for required time
                d = dt.datetime.now()
                diff = (n - d).total_seconds()
                if diff <= 0:
                    # no events should be triggered twice, so if there was
                    # trigger in last 60 seconds, then wait some time
                    if time.time() - lastfire < 60:
                        logger.debug('Wait for 1 minute')
                        # Sleep in smaller chunks to check shutdown flag
                        for _ in range(60):
                            if _is_shutdown_requested():
                                break
                            time.sleep(1)
                    else:
                        lastfire = time.time()
                        logger.debug(f'Trigger event at: {n}')
                        t = threading.Thread(target=func, args=(self, d,))
                        t.daemon = True  # Make thread daemon so it exits when main exits
                        t.start()
                else:
                    waittime = max(min(86400, diff), 1)
                    logger.debug(f'Wait for {waittime} seconds, diff={diff}')
                    # Sleep in smaller chunks to check shutdown flag more frequently
                    sleep_chunks = max(1, int(waittime))
                    sleep_per_chunk = waittime / sleep_chunks
                    for _ in range(sleep_chunks):
                        if _is_shutdown_requested():
                            break
                        time.sleep(sleep_per_chunk)
        
        logger.info('Scheduler shutdown completed gracefully.')

class ScheduledPlanPool:
    PLANS = []
    PLANS_LOCK = threading.Lock()
    PLANS_THREAD = None
    PLANS_THREAD_RUNNING = False

    @staticmethod
    def len():
        with ScheduledPlanPool.PLANS_LOCK:
            return len(ScheduledPlanPool.PLANS)


    @staticmethod
    def get(idx:int):
        with ScheduledPlanPool.PLANS_LOCK:
            if idx < 0 or idx >= len(ScheduledPlanPool.PLANS):
                return (None, None)
            return ScheduledPlanPool.PLANS[idx]

    @staticmethod
    def check_thread():
        with ScheduledPlanPool.PLANS_LOCK:
            if not ScheduledPlanPool.PLANS_THREAD_RUNNING:
                ScheduledPlanPool.PLANS_THREAD_RUNNING = True
                ScheduledPlanPool.PLANS_THREAD = threading.Thread(target=ScheduledPlanPool.runner)
                ScheduledPlanPool.PLANS_THREAD.daemon = True  # Make thread daemon so it exits when main exits
                ScheduledPlanPool.PLANS_THREAD.start()

    @staticmethod
    def runner():
        while not _is_shutdown_requested():
            d = dt.datetime.now()
            for idx in range(ScheduledPlanPool.len()):
                plan, func = ScheduledPlanPool.get(idx)
                if plan is not None and plan.can_run(d):
                    logger.debug(f'Trigger event at: {d}')
                    t = threading.Thread(target=func, args=(plan, d,))
                    t.daemon = True  # Make thread daemon so it exits when main exits
                    t.start()
            time.sleep(ScheduledPlan.g_period)
        
        # Mark thread as not running when exiting
        with ScheduledPlanPool.PLANS_LOCK:
            ScheduledPlanPool.PLANS_THREAD_RUNNING = False
        logger.info('ScheduledPlanPool runner shutdown completed gracefully.')

def run_async_plan(plan: ScheduledPlan, func:callable):
    if is_test_mode():
        print('Subprocess mode detected, scheduler will not be started.')
        return

    ScheduledPlanPool.check_thread()

    with ScheduledPlanPool.PLANS_LOCK:
        ScheduledPlanPool.PLANS.append((plan, func))


#
# Creating decorator for function, that is called in scheduled times
# Automatically starts scheduler in the background
#
def ScheduledRun(script=None, minute=None, hour=None, day=None, month=None,
                 year=None, weekday=None, yearday=None):
    def decorator(func):
        def wrapper(scheduler, current_time):
            return func(scheduler, current_time)
        if not is_test_mode():
            if script is not None:
                plan = ScheduledPlan("plan")
                plan.analyze(script)
                run_async_plan(plan, func)
            else:
                s = SchedulerConfig(minute=minute, hour=hour,
                                    weekday=weekday, yearday=yearday,
                                    day=day, month=month, year=year)
                s.run_async(func)
            time.sleep(1)
        return wrapper
    return decorator

def is_shutdown_requested():
    """
    Public function to check if shutdown has been requested.
    
    Returns:
        bool: True if shutdown has been requested, False otherwise.
    """
    return _is_shutdown_requested()




#!/usr/bin/env python3
"""
Example demonstrating signal handling in the scheduler.

This example shows how to use the new signal handling features:
1. Schedulers will automatically stop on SIGTERM, SIGINT (Ctrl+C), and SIGBREAK (Ctrl+Break on Windows)
2. You can manually request shutdown using request_shutdown()
3. You can use wait_for_shutdown() to block until a signal is received
4. You can check if shutdown was requested using is_shutdown_requested()
"""

import time
import logging
from scheduler_pkg import scheduled, request_shutdown, is_shutdown_requested

# Setup logging to see what's happening
logging.basicConfig(level=logging.INFO, format='%(asctime)s - %(levelname)s - %(message)s')
logger = logging.getLogger(__name__)

# Counter to track how many times the task has run
task_counter = 0

@scheduled(minute='*/1')  # Run every minute
def my_scheduled_task(scheduler, current_time):
    global task_counter
    task_counter += 1
    logger.info(f"Task executed #{task_counter} at {current_time}")
    
    # Simulate some work
    time.sleep(2)
    
    # Example: manually request shutdown after 5 executions
    if task_counter >= 5:
        logger.info("Task has run 5 times, requesting shutdown...")
        request_shutdown()


ftask_counter = 0
ftask_counter2 = 0

@scheduled("at * on * freq 100/hour")
def my_frequent_task(scheduled_plan, current_time):
    global ftask_counter
    ftask_counter += 1
    logger.info(f"Frequent task executed #{ftask_counter} at {current_time}")
    
    # Simulate some work
    time.sleep(1)

@scheduled("at * on * freq 120/hour")
def my_frequent_task2(scheduled_plan, current_time):
    global ftask_counter2
    ftask_counter2 += 1
    logger.info(f"Frequent task2 executed #{ftask_counter2} at {current_time}")
    
    # Simulate some work
    time.sleep(1)

def wait_for_shutdown(check_interval=1):
    """
    Wait for shutdown signal in a blocking manner.
    
    This function can be used in main threads to wait for termination signals
    instead of using infinite loops or sleep. It will return when a shutdown
    signal is received.
    
    Args:
        check_interval (float): How often to check for shutdown signal in seconds.
                               Default is 1 second.
    """
    try:
        while not is_shutdown_requested():
            time.sleep(check_interval)
    except KeyboardInterrupt:
        logger.info("KeyboardInterrupt received, initiating shutdown...")
        request_shutdown()
    
    logger.info("Shutdown signal received, exiting wait_for_shutdown()")


def main():
    logger.info("Starting scheduler example with signal handling...")
    logger.info("Press Ctrl+C to gracefully stop the scheduler")
    logger.info("The scheduler will also stop automatically after the task runs 5 times")
    
    # The @ScheduledRun decorator has already started the scheduler in the background
    
    try:
        # Wait for shutdown signal (blocks until signal received)
        wait_for_shutdown()
    except KeyboardInterrupt:
        logger.info("KeyboardInterrupt caught in main, requesting shutdown...")
        request_shutdown()
    
    # Give some time for background threads to finish
    timeout = 10
    logger.info(f"Waiting up to {timeout} seconds for background tasks to finish...")
    for i in range(timeout):
        time.sleep(1)
        if not is_shutdown_requested():
            break
    
    logger.info(f"Scheduler stopped. Task ran {task_counter} times total.")

if __name__ == "__main__":
    main()
